from . import openapi
from . import generate
import click


@click.command()
@click.option(
    "--package-name",
    help="Name of package to create routes for. Package name is used to create \
import paths, so not mistakes here.",
)
@click.option(
    "--filename", help="OpenAPI filename (JSON file)", default="api-spec.json"
)
def generate_routes(filename, package_name):
    """Create `routes.py` and `test_routes.py` based on OpenAPI file.

    :param filename: json OpenAPI filename
    :type filename: str
    :param package_name: package name
    :type package_name: str
    """
    file = openapi.load_json_file(filename)
    routes = openapi.parse_routes(oa_file=file, package_name=package_name)
    generate.routes(routes, package_name)
    generate.test_routes(routes, package_name)



from minty import Base
from minty.cqrs import CQRS
from minty.infrastructure import InfrastructureFactory
from pyramid.config import Configurator


def _build_cqrs_setup(cqrs):
    """Create a callable for setting up the "CQRS" methods request objects.

    :param cqrs: A configured CQRS object
    :type cqrs: CQRS
    :return: A function, callable by Pyramid, to register the CQRS
        method(s)
    :rtype: callable
    """

    def setup_cqrs_request(config):
        """Add the CQRS accessors to the Pyramid request objects.

        :param config: Pyramid configurator instance
        :type config: Configurator
        :return: Nothing
        :rtype: None
        """

        def get_query_instance(request, domain: str):
            return cqrs.get_query_instance(domain, context=request.host)

        config.add_request_method(get_query_instance, "get_query_instance")

        def get_command_instance(request, domain: str):
            return cqrs.get_command_instance(domain, context=request.host)

        config.add_request_method(get_command_instance, "get_command_instance")

    return setup_cqrs_request


class Engine(Base):
    """Pyramid configurator."""

    __slots__ = ["config", "domains"]

    def __init__(self, domains: list):
        self.domains = domains
        self.config = None

    def setup(self, global_config: dict, **settings) -> object:
        """Set up the application by loading the injecting the CQRS layer.

        :param global_config: Global configuration
        :type global_config: dict
        :return: Returns the Configurator from Pyramid
        :rtype: object
        """
        infra_factory = InfrastructureFactory(
            settings["minty_service.infrastructure.config_file"]
        )

        cqrs = CQRS(self.domains, infra_factory)

        config = Configurator(settings=settings)
        config.include(_build_cqrs_setup(cqrs))

        # TODO 400/404/500 handlers (JSON Output)
        config.add_tween("minty_pyramid.loader.RequestTimer")

        config.scan()
        self.config = config
        return config

    def main(self) -> object:
        """Run the application by calling the wsgi_app function of Pyramid.

        :raises ValueError: When setup is forgotten
        :return: wsgi app
        :rtype: object
        """
        if self.config is None:
            raise ValueError("Make sure you run setup before 'main'")

        self.logger.info("Creating WSGI application")
        return self.config.make_wsgi_app()


class RequestTimer(Base):
    """Middleware / tween to log request time and count to statsd."""

    def __init__(self, handler, registry):
        self.handler = handler
        self.registry = registry

    def __call__(self, request):
        """Log time and count to statsd before and after execution of handler.

        :param request: request
        :type request: class
        :return: response
        :rtype: class
        """
        timer = self.statsd.get_timer()
        timer.start()

        response = self.handler(request)

        try:
            req_name = request.matched_route.name
        except AttributeError:
            req_name = "unknown"

        timer.stop(f"{req_name}.time")

        self.statsd.get_counter().increment(
            f"{req_name}.{response.status_int}.count"
        )

        return response



from jinja2 import Environment, PackageLoader


def load_template(template_name: str):
    """Load template from template folder.

    :param template_name: template name
    :type template_name: str
    :return: template
    :rtype: jinja2 template
    """
    pkg_loader = PackageLoader(__name__)
    env = Environment(loader=pkg_loader)
    return env.get_template(template_name)


def routes(routes: list, package_name: str):
    """Create `routes.py` for given routes in package.

    :param routes: routes to create file with
    :type routes: list
    :param package_name: package name
    :type package_name: str
    :return: file created
    :rtype: bool
    """
    template = load_template("routes.j2")
    content = template.render(routes=routes)
    return create_file(
        content=content, filename="routes.py", path=f"{package_name}/routes"
    )


def test_routes(routes: list, package_name):
    """Create 'test_routes_routes.py' for given routes in package.

    :param routes: routes to create file with
    :type routes: list
    :param package_name: package name
    :type package_name: str
    :return: file created
    :rtype: bool
    """
    grouped_routes = group_routes_by(routes=routes, param="handler_file")
    t = load_template("test_routes.j2")
    content = t.render(routes=grouped_routes, package_name=package_name)
    return create_file(
        content=content, filename="test_routes.py", path="tests"
    )


def group_routes_by(routes: list, param: str) -> dict:
    """Group routes by parameter.

    Create different groupings to accomodate different templates

    :param routes: routes
    :type routes: list
    :param param: parameter to group by
    :type param: str
    :return: grouped routes
    :rtype: dict
    """
    grouped_routes = {}
    for route in routes:
        grouped_by_param = grouped_routes.get(route[param], [])
        grouped_by_param.append(route)
        grouped_routes[route[param]] = grouped_by_param
    return grouped_routes


def create_file(content, filename: str, path: str):
    """Create file on given location and write content to it.

    :param content: content to write to file
    :type content: str
    :param filename: filename
    :type filename: str
    :param path: path to file
    :type path: str
    """
    filepath = f"{path}/{filename}"

    with open(filepath, "w", encoding="utf-8") as file:
        file.write(content)
        print(f"Created: '{filepath}'")



import json


def parse_routes(oa_file, package_name: str) -> list:
    """Parse OpenAPI file and return a list of routes.

    this function expects the following directory structure in the application:
    `{package_name}/views/{x-view}`
    where `x-view` is the filename for the handlers specified in the OpenAPI
    file.

    :param oa_file: OpenAPI dict
    :type oa_file: dict
    :param package_name: package name
    :type package_name: str
    :return: list of routes
    :rtype: list
    """
    routes = []
    for route in oa_file["paths"]:
        r = oa_file["paths"][route]
        for method in r:
            h_file = r[method]["x-view"]
            handler = r[method]["operationId"]

            out = {
                "route": route,
                "method": method.upper(),
                "handler": handler,
                "renderer": "json",
                "view": f"{package_name}.views.{h_file}.{handler}",
                "handler_file": h_file,
            }
            routes.append(out)
    return routes


def load_json_file(filename: str) -> dict:
    """Load an OpenAPI JSON file.

    :param filename: filename
    :type filename: str
    :return: OpenAPI file
    :rtype: dict
    """
    with open(filename, "r", encoding="utf-8") as jsonfile:
        return json.load(jsonfile)
